//+------------------------------------------------------------------+
//|                                                          dtw.mqh |
//|                                  Copyright 2024, MetaQuotes Ltd. |
//|                                             https://www.mql5.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2024, MetaQuotes Ltd."
#property link      "https://www.mql5.com"
#include<np.mqh>
//+------------------------------------------------------------------+
//|  step patterns                                                   |
//+------------------------------------------------------------------+
enum ENUM_STEP_PATTERN
  {
   STEP_SYMM1=0,//symmetric1
   STEP_SYMM2,//symmetric2
   STEP_ASYMM//asymmetric
  };
//+------------------------------------------------------------------+
//| hint                                                             |
//+------------------------------------------------------------------+
enum ENUM_HINT
  {
   HINT_NA=0,
   HINT_N,
   HINT_M,
   HINT_NM
  };

//+------------------------------------------------------------------+
//| lists the transitions allowed while searching                    |
//|for the minimum-distance path                                     |
//+------------------------------------------------------------------+
class CStepPattern
  {
private:
   matrix            m_mx,m_stepsMatrix;
   ENUM_HINT         m_stephint;
public:
                     CStepPattern(matrix &mx, matrix &stepmatrix, ENUM_HINT hint=HINT_NA)
     {
      m_mx = mx;
      m_stepsMatrix = stepmatrix;
      m_stephint = hint;
     }
                    ~CStepPattern(void)
     {

     }
   matrix            getStepMatrix(void)
     {
      return m_stepsMatrix;
     }
   ulong             getNRows(void)
     {
      return m_mx.Rows();
     }
   int               getNPatterns(void)
     {
      vector vec = m_mx.Col(0);
      return (int(vec.Max()));
     }

   CStepPattern*     T(void)
     {
      ulong cols[] = {0, 2, 1, 3};

      matrix cpy = np::selectMatrixCols(m_mx,cols);
      ENUM_HINT hint = m_stephint;

      if(m_stephint == HINT_N)
         hint = HINT_M;
      else
         if(m_stephint == HINT_M)
            hint = HINT_N;

      CStepPattern* out = new CStepPattern(cpy,m_stepsMatrix,hint);

      return out;
     }

   matrix            extractPattern(vector &sn)
     {
      vector col  =  m_mx.Col(0);
      matrix out = matrix::Ones(1,1);
      for(ulong i = 0; i<m_mx.Rows(); i++)
        {
         for(ulong j = 0; j<col.Size(); j++)
           {
            if(col[j] == sn[j])
              {
               if(!out.Resize(out.Rows()+1,m_mx.Cols()-1,100))
                 {
                  Print(__FUNCTION__, " error ", GetLastError());
                  return matrix::Zeros(1,1);
                 }
               vector v = m_mx.Row(j);
               vector vv = np::sliceVector(v,1);
               if(!out.Row(vv,out.Rows()-1))
                 {
                  Print(__FUNCTION__, " error ", GetLastError());
                  return matrix::Zeros(1,1);
                 }
              }
           }
        }

      if(!np::reverseMatrixRows(out))
        {
         Print(__FUNCTION__, " Reverse Matrix failure ");
         return matrix::Zeros(1,1);
        }
      return out;
     }

   matrix            mkDIrDeltas(void)
     {
      matrix out = matrix::Zeros(1,1);
      vector col = m_mx.Col(3);
      for(ulong i = 0; i<m_mx.Rows(); i++)
        {
         for(ulong j = 0; j<col.Size(); j++)
           {
            if(col[j] == -1.0)
              {
               if(!out.Resize(out.Rows()+1,m_mx.Cols(),100))
                 {
                  Print(__FUNCTION__, " error ", GetLastError());
                  return matrix::Zeros(1,1);
                 }
               vector v = m_mx.Row(j);
               if(!out.Row(v,out.Rows()-1))
                 {
                  Print(__FUNCTION__, " error ", GetLastError());
                  return matrix::Zeros(1,1);
                 }
              }
           }
        }

      return np::sliceMatrixCols(out,1,3);
     }

   matrix             getP(void)
     {
      ulong sel[] = {0, 2, 1, 3};
      matrix s = np::selectMatrixCols(m_mx,sel);
      return s;
     }

   matrix            getMx(void)
     {
      return m_mx;
     }
  };
//---
matrix _symmetric1 =
  {
     {1, 1, 1, -1},
     {1, 0, 0, 1},
     {2, 0, 1, -1},
     {2, 0, 0, 1},
     {3, 1, 0, -1},
     {3, 0, 0, 1}
  };


matrix _symmetric2 =
  {
     {1, 1, 1, -1},
     {1, 0, 0, 2},
     {2, 0, 1, -1},
     {2, 0, 0, 1},
     {3, 1, 0, -1},
     {3, 0, 0, 1}
  };


matrix _asymmetric =
  {
     {1, 1, 0, -1},
     {1, 0, 0, 1},
     {2, 1, 1, -1},
     {2, 0, 0, 1},
     {3, 1, 2, -1},
     {3, 0, 0, 1}
  };

matrix Symmetric =
  {
     {1, 1},
     {0, 1},
     {1, 0}
  };


matrix Asymmetric =
  {
     {1, 0},
     {1, 1},
     {1, 2}
  };
//+------------------------------------------------------------------+
//|  distance metric                                                 |
//+------------------------------------------------------------------+
enum ENUM_DIST_METRIC
  {
   DIST_EUCLIDEAN=0,//euclidean
   DIST_CITYBLOCK,//city block
   DIST_COSINE,//cosine
   DIST_CORRELATION,//correlation
   DIST_CHEBYSHEV,//chebyshev
   DIST_SQEUCLIDEAN//squared euclidean
  };

//+------------------------------------------------------------------+
//|   window function                                                |
//+------------------------------------------------------------------+
enum ENUM_GLOBAL_CONSTRAINT
  {
   CONSTRAINT_NONE=0,// no constrains applied
   CONSTRAINT_SAKOECHIBA,// sakoe chiba
   CONSTRAINT_SLATEDBAND,// slated band
   CONSTRAINT_ITAKURA// itakura
  };
//+------------------------------------------------------------------+
//| Global constraints                                               |
//+------------------------------------------------------------------+
class CConstraint
  {
public:
                     CConstraint(void)
     {

     }
                    ~CConstraint(void)
     {

     }

   static matrix       noConstraint(ulong iw,ulong jw)
     {
      matrix mats[];

      np::indices(iw,jw,mats);

      for(ulong i = 0; i<mats[0].Rows(); i++)
        {
         for(ulong j = 0; j<mats[0].Cols(); j++)
           {
            long value = long(mats[0][i][j]);
            mats[0][i][j] = (double)(value|1);
            if(mats[0][i][j]==0.0)
               mats[0][i][j] = double("inf");
           }
        }

      return mats[0];
     }

   static matrix       sakoeChibaConstraint(ulong iw,ulong jw, ulong qsize, ulong refsize, ulong winsize)
     {
      matrix mats[];

      np::indices(iw,jw,mats);

      matrix abs = MathAbs(mats[1]-mats[0]);

      for(ulong i = 0; i<abs.Rows(); i++)
        {
         for(ulong j = 0; j<abs.Cols(); j++)
           {
            if(ulong(abs[i][j])<=winsize)
               abs[i][j] = (double)(1);
            else
               abs[i][j] = double("inf");
           }
        }

      return abs;
     }

   static matrix       itakuraConstraint(ulong iw,ulong jw, ulong qsize, ulong refsize)
     {
      matrix mats[];

      np::indices(iw,jw,mats);

      long a,b,c,d;

      for(ulong i = 0, k = 0; i<mats[0].Rows() && k<mats[1].Rows(); i++,k++)
        {
         for(ulong j = 0; j<mats[0].Cols(); j++)
           {
            a = long(mats[1][k][j]) < (2*long(mats[0][i][j]))?1:0;
            b = long(mats[0][i][j]) <=(2*long(mats[1][k][j]))?1:0;
            c = long(mats[0][i][j]) >=(long(qsize)-1-2*(long(refsize)-long(mats[1][k][j])))?1:0;
            d = long(mats[1][k][j]) > (long(refsize)-1-2*(long(qsize)-long(mats[0][i][j])))?1:0;
            mats[0][i][j] = double(ulong(a&b&c&d));
            if(mats[0][i][j]==0.0)
               mats[0][i][j] = double("inf");
           }
        }

      return mats[0];
     }

   static matrix       slantedBandConstraint(ulong iw,ulong jw, ulong qsize, ulong refsize,ulong winsize)
     {

      matrix mats[];

      np::indices(iw,jw,mats);

      matrix  diagj = (mats[0]*refsize/qsize);

      matrix abs = MathAbs(mats[1]-diagj);


      for(ulong i = 0; i<abs.Rows(); i++)
        {
         for(ulong j = 0; j<abs.Cols(); j++)
           {
            if(ulong(abs[i][j])<=winsize)
               abs[i][j] = (double)(1);
            else
               abs[i][j] = double("inf");
           }
        }

      return abs;

     }

  };
//+------------------------------------------------------------------+
//| Main dynamic time warping class                                  |
//+------------------------------------------------------------------+
class Cdtw
  {
private:
   matrix            m_distance,m_dirMatrix,m_costMatrix,m_warpPath,m_query,m_ref;
   ulong             m_qlen,m_reflen,m_jmin,m_winsize,m_np;
   double            m_dist,m_ndistance;
   CStepPattern *    m_stepPattern;
   ENUM_DIST_METRIC  m_distMetric;
   ENUM_GLOBAL_CONSTRAINT m_winMethod;
   ENUM_HINT         m_stephint;
   //+------------------------------------------------------------------+
   //|  private method implementing accumulated cost calculation        |
   //+------------------------------------------------------------------+

   bool              calCM(matrix &distMatrix,matrix &stepMatrix,matrix &costMatrix,matrix &dirMatrix)
     {
      ulong max0,max1;
      max0 = ulong(stepMatrix.Col(0).Max());
      max1 = ulong(stepMatrix.Col(1).Max());
      double curCost,curd;

      for(ulong i = max0; i<costMatrix.Rows(); i++)
        {
         for(ulong j = max1; j<costMatrix.Cols(); j++)
           {
            for(ulong k = 0; k<stepMatrix.Rows(); k++)
              {
               curd = costMatrix[i-ulong(stepMatrix[k][0])][j-ulong(stepMatrix[k][1])];
               curCost = curd + distMatrix[i-max0][j-max1];
               if(curCost<costMatrix[i][j])
                 {
                  costMatrix[i][j] = curCost;
                  dirMatrix[i][j] = double(k);
                 }
              }
           }
        }
      costMatrix = np::sliceMatrix(costMatrix,max0,END,1,max1);
      dirMatrix  = np::sliceMatrix(dirMatrix,max0,END,1,max1);
      return true;

     }
   //+------------------------------------------------------------------+
   //|  general minkowski distance metric                               |
   //+------------------------------------------------------------------+


   double            minkowski(vector &u,vector &v, int p = 2)
     {
      if(p<=0)
        {
         Print(__FUNCTION__, " p must be greater than 0");
         return EMPTY_VALUE;
        }
      vector uv = u - v;
      return uv.Norm(VECTOR_NORM_P,p);
     }
   //+------------------------------------------------------------------+
   //|  euclidean distance calculation                                  |
   //+------------------------------------------------------------------+

   double            euclidean(vector &u,vector &v)
     {
      return minkowski(u,v);
     }

   //+------------------------------------------------------------------+
   //|    squared euclidean distance                                    |
   //+------------------------------------------------------------------+


   double            sqeuclidean(vector &u,vector &v)
     {
      vector uv = u - v;
      vector uvw = uv;
      return uv.Dot(uvw);
     }
   //+------------------------------------------------------------------+
   //|  correlation distance                                            |
   //+------------------------------------------------------------------+

   double            correlation(vector &u,vector &v, bool centred=true)
     {
      double umu,vmu;
      
      if(centred)
        {
            umu=u.Mean();
            vmu=v.Mean();
           
         u = u - umu;
         v = v - vmu;
        }
      vector vw,uw;
      
         vw = v;
         uw = u;
         
      double uv,uu,vv,dist;
      uv = u.Dot(vw);
      uu = u.Dot(uw);
      vv = v.Dot(vw);
      dist = 1.0 - uv/sqrt(uu*vv);

      if(dist<0.0)
         dist = 0.0;
      else
         if(dist>2.0)
            dist = 2.0;

      return dist;

     }
   //+------------------------------------------------------------------+
   //|   cosine distance                                                |
   //+------------------------------------------------------------------+

   double            cosine(vector &u,vector &v)
     {
      return correlation(u,v,false);
     }
   //+------------------------------------------------------------------+
   //|   city block distance                                            |
   //+------------------------------------------------------------------+

   double            cityblock(vector &u,vector &v)
     {
      vector diff = MathAbs(u-v);
      return diff.Sum();
     }
   //+------------------------------------------------------------------+
   //|    chebyshev distance                                            |
   //+------------------------------------------------------------------+

   double            chebyshev(vector &u,vector &v)
     {
      return (MathAbs(u-v)).Max();
     }

   //+------------------------------------------------------------------+
   //|  distance metric calculation                                     |
   //+------------------------------------------------------------------+


   double            dist(vector &u,vector &v)
     {
      switch(m_distMetric)
        {
         case DIST_EUCLIDEAN:
            return euclidean(u,v);
         case DIST_CITYBLOCK:
            return cityblock(u,v);
         case DIST_CHEBYSHEV:
            return chebyshev(u,v);
         case DIST_CORRELATION:
            return correlation(u,v);
         case DIST_COSINE:
            return cosine(u,v);
         case DIST_SQEUCLIDEAN:
            return sqeuclidean(u,v);
         default:
            Print(__FUNCTION__, " invalid parameter ");
            return EMPTY_VALUE;
        }
     }

   //+------------------------------------------------------------------+
   //|  optimum path search function                                    |
   //+------------------------------------------------------------------+

   matrix            backtrack(matrix &dirMatrix, matrix &stepMatrix, bool subseq,long startFrom)
     {
      vector curid(2);

      if(startFrom<0)
        {
         curid[0] = double(dirMatrix.Rows() - 1);
         curid[1] = double(dirMatrix.Cols() - 1);
        }
      else
        {
         curid[0] = double(dirMatrix.Rows() - 1);
         curid[1] = double(startFrom);
        }

      matrix out(1,2);
      out.Row(curid,0);
      double step;
      while((subseq && curid[0] > 0.0) || (!subseq && curid[0] != 0.0))
        {
         step = dirMatrix[ulong(curid[0])][ulong(curid[1])];
         curid[0] = curid[0] - stepMatrix[ulong(step)][0];
         curid[1] = curid[1] - stepMatrix[ulong(step)][1];
         if(curid.Min() < 0.0)
            break;

         if(!out.Resize(out.Rows()+1,out.Cols(),100) || !out.Row(curid,out.Rows()-1))
           {
            Print(__FUNCTION__," row insertion error ", GetLastError());
            break;
           }

        }

      return out;
     }


public:
   //+------------------------------------------------------------------+
   //|  default constructor                                             |
   //+------------------------------------------------------------------+

                     Cdtw(void)
     {

     }
   //+------------------------------------------------------------------+
   //|    destructor                                                    |
   //+------------------------------------------------------------------+

                    ~Cdtw(void)
     {
      if(CheckPointer(m_stepPattern)==POINTER_DYNAMIC)
         delete m_stepPattern;
     }
   //+------------------------------------------------------------------+
   //|     main interface method for dtw                                |
   //+------------------------------------------------------------------+

   bool              dtw(matrix&x, matrix&y, ENUM_DIST_METRIC dist_method,ENUM_STEP_PATTERN step_pattern=STEP_SYMM2, ENUM_GLOBAL_CONSTRAINT win_type=CONSTRAINT_NONE,ulong winsize=0)
     {
      if(y.Cols()!=x.Cols())
        {
         Print(__FUNCTION__, " invalid input parameters, size containers donot match. ");
         return false;
        }

      if(CheckPointer(m_stepPattern)==POINTER_DYNAMIC)
         delete m_stepPattern;

      switch(step_pattern)
        {
         case STEP_SYMM1:
            m_stepPattern = new CStepPattern(_symmetric1,Symmetric);
            m_stephint = HINT_NA;
            break;

         case STEP_SYMM2:
            m_stepPattern = new CStepPattern(_symmetric2,Symmetric,HINT_NM);
            m_stephint = HINT_NM;
            break;

         case STEP_ASYMM:
            m_stepPattern = new CStepPattern(_asymmetric,Asymmetric,HINT_N);
            m_stephint = HINT_N;
            break;
        }

      if(CheckPointer(m_stepPattern)==POINTER_INVALID)
        {
         Print(__FUNCTION__," failed step pointer initialization ", GetLastError());
         return false;
        }

      matrix stepsMatrix = m_stepPattern.getStepMatrix();
      m_query = x;
      m_qlen = x.Rows();
      m_ref = y;
      m_reflen = y.Rows();
      m_distMetric = dist_method;
      m_winMethod = win_type;
      m_winsize = winsize;

      if(y.Rows())
        {
         if(!m_distance.Resize(m_qlen,m_reflen))
           {
            Print(__FUNCTION__," resize error ", GetLastError());
            return false;
           }
         for(ulong i = 0; i<m_qlen; i++)
            for(ulong j =0; j<m_reflen; j++)
               m_distance[i][j]=dist(m_query.Row(i),m_ref.Row(j));
        }
      else
         m_distance = m_query;

      ulong n,m;
      n=m_distance.Rows();
      m=m_distance.Cols();
      matrix wm;
      if(m_winMethod == CONSTRAINT_NONE)
         wm = matrix::Ones(m_distance.Rows(), m_distance.Cols());
      else
        {
         switch(m_winMethod)
           {
            case CONSTRAINT_ITAKURA:
               wm = CConstraint::itakuraConstraint(n,m,m_qlen,m_reflen);
               break;
            case CONSTRAINT_SAKOECHIBA:
               wm = CConstraint::sakoeChibaConstraint(n,m,m_qlen,m_reflen,m_winsize);
               break;
            case CONSTRAINT_SLATEDBAND:
               wm = CConstraint::slantedBandConstraint(n,m,m_qlen,m_reflen,m_winsize);
               break;
            default:
               wm = CConstraint::noConstraint(n,m);
               break;
           }
        }

      if(m_winMethod!=CONSTRAINT_NONE)
        {
         for(ulong i = 0; i<wm.Rows(); i++)
            for(ulong j = 0; j<wm.Cols(); j++)
               if((i+j)>0 && wm[i][j] != 1.0)
                  m_distance[i][j] = wm[i][j];
        }


      m_costMatrix = matrix::Zeros(m_distance.Rows()+ulong(stepsMatrix.Col(0).Max()),m_distance.Cols()+ulong(stepsMatrix.Col(1).Max()));
      m_costMatrix.Fill(double("inf"));
      m_costMatrix[ulong(stepsMatrix.Col(0).Max())][ulong(stepsMatrix.Col(1).Max())] = m_distance[0][0];

      m_dirMatrix = matrix::Zeros(m_costMatrix.Rows(),m_costMatrix.Cols());
      m_dirMatrix.Fill(double(INT_MIN));
      for(ulong i = 0; i<m_dirMatrix.Cols(); i++)
         m_dirMatrix[0][i] = double(1);
      for(ulong i = 0; i<m_dirMatrix.Rows(); i++)
         m_dirMatrix[i][0] = double(2);

      if(!calCM(m_distance,stepsMatrix,m_costMatrix,m_dirMatrix))
        {
         Print(__FUNCTION__, " computeCM() failed ");
         return false;
        }

      m_jmin = m_costMatrix.Cols() - 1;

      return true;
     }
   //+------------------------------------------------------------------+
   //| overloaded implementation for vectors (univariate time series)   |
   //+------------------------------------------------------------------+

   bool              dtw(vector&x, vector&y,ENUM_DIST_METRIC dist_method,ENUM_STEP_PATTERN step_pattern=STEP_SYMM2, ENUM_GLOBAL_CONSTRAINT win_type=CONSTRAINT_NONE,ulong winsize=0)
     {
      matrix query(x.Size(),1);
      matrix ref(y.Size(),1);
      if(!query.Col(x,0) || !ref.Col(y,0))
        {
         Print(__FUNCTION__," column insertion failed ", GetLastError());
         return false;
        }
      return dtw(query,ref,dist_method,step_pattern,win_type,winsize);
     }
   //+------------------------------------------------------------------+
   //|    Get the optimal path: corresponding points from both series   |
   //+------------------------------------------------------------------+

   matrix            warpPath(bool openEnd=false)
     {
      matrix stmatrix = m_stepPattern.getStepMatrix();
      return backtrack(m_dirMatrix,stmatrix,openEnd,openEnd?long(m_costMatrix.Row(m_costMatrix.Rows()-1).ArgMin()):-1);
     }
   //+------------------------------------------------------------------+
   //|     Get the accumulated cost matrix                              |
   //+------------------------------------------------------------------+

   matrix            costMatrix(void)
     {
      return m_costMatrix;
     }
   //+------------------------------------------------------------------+
   //|       Get the cost matrix                                        |
   //+------------------------------------------------------------------+

   matrix            localCostMatrix(void)
     {
      return m_distance;
     }
   //+------------------------------------------------------------------+
   //|        Get the direction matrix                                  |
   //+------------------------------------------------------------------+

   matrix            directionMatrix(void)
     {
      return m_dirMatrix;
     }
   //+------------------------------------------------------------------+
   //|         Get the normalized distance                              |
   //+------------------------------------------------------------------+

   double            normalizedDistance(void)
     {
       vector lastrow = m_costMatrix.Row(m_costMatrix.Rows()-1);

      if(m_stephint == HINT_NM)
        {
         vector v = np::arange(m_costMatrix.Cols());
         lastrow = lastrow/(double(m_costMatrix.Rows()) + v + 1.0);
        }
      else
         if(m_stephint == HINT_N)
            lastrow = lastrow/double(m_costMatrix.Rows());
         else
            if(m_stephint == HINT_M)
              {
               vector v =  np::arange(m_costMatrix.Cols());
               lastrow = lastrow/(1.0+v);
              }
              
      if(m_stephint != HINT_NA)
         return lastrow[m_jmin];
      else
         return double("nan");
     }
   //+------------------------------------------------------------------+
   //|        Get the distance                                          |
   //+------------------------------------------------------------------+

   double            distance(void)
     {
      return m_costMatrix[m_costMatrix.Rows()-1][m_jmin];
     }


  };
//+------------------------------------------------------------------+



